from yass.dlx.models import Column, Node


def uncover(col: Column):
    """Include the Column and associated rows in the search."""
    assert col.left and col.right
    cover_node = col.up
    while cover_node and cover_node != col:
        node = cover_node.left
        while isinstance(node, Node) and node and node != cover_node:
            assert node.up and node.down and node.col
            node.up.down = node
            node.down.up = node
            node.col.size += 1
            node = node.left
        cover_node = cover_node.up
    col.left.right = col
    col.right.left = col



from typing import Iterator

from yass.models import Board, Solver


def candidate_boards(board: Board, addr: str, depth: int = 0) -> Iterator[Board]:
    candidates = board.candidates(addr)
    assert candidates is not None
    for candidate in candidates:
        board.set(addr, candidate)
        yield board
        naddr = board.next(addr)
        yield from candidate_boards(board, naddr, depth=depth + 1)
        board.unset(addr)


class Backtrack:
    def solve(self, board: Board, **kwargs) -> Board:
        """Solve a sudoku puzzle."""

        for board in candidate_boards(board, "11"):
            if board.is_completed():
                break
        return board



from yass.dlx.models import Column, Node


def cover(col: Column):
    """Exclude the Column and associated rows from the search."""
    assert col.left and col.right
    col.left.right = col.right
    col.right.left = col.left
    if col.down == col:
        return
    cover_node = col.down
    while cover_node and cover_node != col:
        node = cover_node.right
        while isinstance(node, Node) and node and node != cover_node:
            assert node.up and node.down and node.col
            node.up.down = node.down
            node.down.up = node.up
            node.col.size -= 1
            node = node.right
        cover_node = cover_node.down



from __future__ import annotations
from dataclasses import dataclass
from typing import Protocol


BOARDS = [
    (
        "53..7...."
        "6..195..."
        ".98....6."
        "8...6...3"
        "4..8.3..1"
        "7...2...6"
        ".6....28."
        "...419..5"
        "....8..79"
    ),
]


class Solver(Protocol):
    def solve(self, board: Board, **kwargs) -> Board:
        ...


def box(col, row):
    return (col - 1) // 3 + 3 * ((row - 1) // 3) + 1


@dataclass(slots=True)
class Cell:
    col: str
    row: str
    box: str
    val: str | None = None
    og: bool = False

    def __str__(self):
        return f"{self.val or '.'}"

    def with_val(self, val: str) -> Cell:
        if self.is_set():
            raise ValueError(
                f"(col={self.col}, row={self.row}) is set, not setting to {val}"
            )
        return Cell(col=self.col, row=self.row, box=self.box, val=val, og=self.og)

    def neighbour(self, c: Cell) -> bool:
        if self.col == c.col and self.row == c.row and self.box == c.box:
            return False
        return (self.col == c.col) or (self.row == c.row) or (self.box == c.box)

    def is_set(self):
        return self.val is not None

    def can_unset(self):
        return not self.og


class InvalidBoard(Exception):
    """"""


def invalid_board(string: str):
    err = {}
    if not len(string) == 81:
        err["msg"] = "Incorrect number of cells"
    return err


class Board:
    """Sudoku board."""

    SIZE = 9
    OPTIONS = "123456789"

    @classmethod
    def from_string(cls, string: str) -> Board:
        board = cls()
        if err := invalid_board(string):
            raise InvalidBoard(err["msg"])
        for addr, val in zip(board, string):
            if val != ".":
                board.set(addr, val)
                board[addr].og = True
        return board

    def __init__(self):
        self._set_count = 0
        self._cells = {
            f"{col}{row}": Cell(
                col=str(col),
                row=str(row),
                box=str(box(col, row)),
            )
            for row in range(1, self.SIZE + 1)
            for col in range(1, self.SIZE + 1)
        }
        self._neighbours = {
            f"{col}{row}": [
                c for _, c in self.items() if c.neighbour(self[f"{col}{row}"])
            ]
            for row in range(1, self.SIZE + 1)
            for col in range(1, self.SIZE + 1)
        }

    def __iter__(self):
        return iter(self._cells)

    def __getitem__(self, k) -> Cell:
        return self._cells[k]

    def __str__(self):
        s = ""
        for row in range(1, self.SIZE + 1):
            for col in range(1, self.SIZE + 1):
                s += str(self[f"{col}{row}"])
                if col in [3, 6]:
                    s += "|"
            s += "\n"
            if row in [3, 6]:
                s += "---|---|---\n"
        return s

    def items(self):
        return self._cells.items()

    def neighbour_vals(self, addr):
        return set(c.val for c in self._neighbours[addr] if c.is_set())

    def candidates(self, addr: str) -> str | None:
        if self[addr].is_set():
            return self[addr].val
        return "".join(set(self.OPTIONS) - self.neighbour_vals(addr))

    def next(self, addr: str) -> str:
        col, row = [int(x) for x in addr]
        if col < self.SIZE:
            return f"{col + 1}{row}"
        return f"1{row + 1}"

    def set(self, addr: str, val: str) -> None:
        if not self[addr].is_set():
            self._set_count += 1
            self[addr].val = val

    def unset(self, addr: str) -> None:
        if self[addr].can_unset():
            self._set_count -= 1
            self[addr].val = None

    def is_completed(self) -> bool:
        return self._set_count == 81



from typing import Sequence

from yass.dlx.models import Column, from_matrix
from yass.dlx.search import search
from yass.models import Board, Cell


def from_board(
    board: Board,
) -> tuple[Column, Sequence[Sequence[int]], tuple[str, ...]]:
    matrix, column_names = remove_empty_cols(*to_matrix(board))
    return (
        from_matrix(matrix=matrix, column_names=column_names),
        matrix,
        column_names,
    )


def cell_to_idx() -> dict[str, int]:
    mapper = {}
    i = 0
    for c in "PRCB":
        for x in range(1, 10):
            for y in range(1, 10):
                mapper[f"{c}{x}{y}"] = i
                i += 1
    return mapper


def cell_to_row(
    cell: Cell, cell_to_idx_mapper: dict[str, int], row_len: int
) -> tuple[int, ...]:
    row = [0] * row_len
    row[cell_to_idx_mapper[f"P{cell.row}{cell.col}"]] = 1
    row[cell_to_idx_mapper[f"R{cell.val}{cell.row}"]] = 1
    row[cell_to_idx_mapper[f"C{cell.val}{cell.col}"]] = 1
    row[cell_to_idx_mapper[f"B{cell.val}{cell.box}"]] = 1
    return tuple(row)


def to_matrix(board: Board) -> tuple[Sequence[Sequence[int]], tuple[str, ...]]:
    cell_to_idx_mapper = cell_to_idx()
    row_len = len(cell_to_idx_mapper)
    rows = []
    for addr, cell in board.items():
        if cell.is_set():
            rows.append(cell_to_row(cell, cell_to_idx_mapper, row_len))
        else:
            candidates = board.candidates(addr)
            assert candidates
            for val in candidates:
                rows.append(
                    cell_to_row(board[addr].with_val(val), cell_to_idx_mapper, row_len)
                )
    return tuple(rows), tuple(cell_to_idx_mapper)


def remove_empty_cols(
    matrix: Sequence[Sequence[int]], col_names: tuple[str, ...]
) -> tuple[Sequence[Sequence[int]], tuple[str, ...]]:
    tr = list(map(list, zip(*matrix)))
    tr_new = []
    col_names_new = []
    for col, col_name in zip(tr, col_names):
        if any(col):
            tr_new.append(col)
            col_names_new.append(col_name)
    return tuple(map(list, zip(*tr_new))), tuple(col_names_new)


def populate_board(
    board: Board, matrix: Sequence[Sequence[int]], col_names: tuple[str, ...]
) -> Board:

    for row in matrix:
        addr_found = False
        for col, name in zip(row, col_names):
            if col:
                if not addr_found:
                    r, c = list(name[1:])  # Remove P
                    addr = f"{c}{r}"  # board addresses are col, row
                    addr_found = True
                else:
                    val = name[1]  # {R,C,B}<val><row,col,box idx>
        board.set(addr, val)
    return board


class Dlx:
    def solve(self, board: Board, **kwargs) -> Board:
        """Solve a sudoku puzzle."""

        col, m, c = from_board(board)
        soln = search(col, soln_length=81)
        soln_matrix = [x for i, x in enumerate(m) if i in soln]
        return populate_board(board=board, matrix=soln_matrix, col_names=c)



import argparse
import logging

from yass.models import Board, Solver
from yass.solve import Backtrack, Dlx


parser = argparse.ArgumentParser()
parser.add_argument("method")
parser.add_argument("board_string")
parser.add_argument("--log_level", default="CRITICAL")


def cli():
    args = parser.parse_args()

    logging.basicConfig(level=args.log_level)

    solvers: dict[str, Solver] = {"backtrack": Backtrack(), "dlx": Dlx()}

    board = Board.from_string(args.board_string)
    print(board)
    solved_board = solvers[args.method].solve(board=board)
    print(solved_board)


if __name__ == "__main__":
    cli()



from yass.dlx.cover import cover
from yass.dlx.models import Column, Node, choose_column
from yass.dlx.uncover import uncover


def search(
    root: Column,
    depth: int = 0,
    soln: list[int] | None = None,
    soln_length: int | None = None,
):
    """Recursive algorithm for exact cover problem."""

    soln = [] if soln is None else soln

    assert root.right

    if root.right.name == "__root__":
        return soln

    col = choose_column(root)

    cover(col)
    down = col.down
    while down and down != col and isinstance(down, Node):
        if not soln or depth >= len(soln):
            soln.append(down.row_idx)
        else:
            soln[depth] = down.row_idx
        right = down.right
        while right and right != down:
            cover(right.col)
            right = right.right
        search(root=root, depth=depth + 1, soln=soln, soln_length=soln_length)
        if soln_length is not None and len(soln) == soln_length:
            return soln
        left = down.left
        while left and left != down:
            uncover(left.col)
            left = left.left
        down = down.down
    uncover(col)
    return soln



from __future__ import annotations
from dataclasses import dataclass
from typing import Sequence


@dataclass
class Node:

    row_idx: int
    col: Column
    left: Node | None = None
    right: Node | None = None
    up: Node | Column | None = None
    down: Node | Column | None = None


@dataclass
class Column:

    name: str
    left: Column | None = None
    right: Column | None = None
    up: Node | Column | None = None
    down: Node | Column | None = None
    size: int = 0


def choose_column(root: Column) -> Column:
    """Choose the next Column object to cover."""
    col = root.right
    min_col = root

    min_size = 1_000_000_000
    while col and col != root:
        s = col.size
        if s < min_size:
            min_col = col
            min_size = s
        col = col.right
    return min_col


def from_matrix(
    matrix: Sequence[Sequence[int]],
    column_names: Sequence[str] | None = None,
) -> Column:
    """Build a Problem object from a sequence of rows."""

    root = Column("__root__")

    cols: list[Column] = []
    if column_names is None:
        ncols: Sequence[str] = [str(i) for i in range(len(matrix[0]))]
    else:
        ncols = column_names

    for i, row in enumerate(matrix):
        if i == 1:
            cols[0].left = root
            cols[-1].right = root
            root.left = cols[-1]
            root.right = cols[0]
        row_nodes: list[Node] = []
        for j, elem in enumerate(row):
            if i == 0:
                col = Column(name=ncols[j])
                if cols:
                    col.left = cols[-1]
                    cols[-1].right = col
                cols.append(col)
            if elem:
                node = Node(
                    row_idx=i,
                    col=cols[j],
                )
                cols[j].size += 1
                up: Node | Column = node.col
                while up.down:
                    up = up.down
                node.up = up
                up.down = node
                if row_nodes:
                    node.left = row_nodes[-1]
                    row_nodes[-1].right = node
                row_nodes.append(node)
        row_nodes[0].left = row_nodes[-1]
        row_nodes[-1].right = row_nodes[0]

    # point all the bottom nodes back to the top nodes and vice versa
    for col in cols:
        bottom = col.down
        while bottom and bottom.down:
            bottom = bottom.down
        assert bottom
        bottom.down = col
        col.up = bottom

    return root
